/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#ifndef _INITIATOR_SOCKET_HPP_
#define _INITIATOR_SOCKET_HPP_

#include "basic.hpp"
#include "ecounters_ext.hpp"
#include "untimed_ext.hpp"
#include "host_memory.hpp"

namespace basic {

  template <typename MODULE, bool MULTIPORT = false>
  class initiator_socket:
    public tlm::tlm_initiator_socket<32,
                                     tlm::tlm_base_protocol_types,
                                     MULTIPORT ? 0 : 1>,
    private tlm::tlm_bw_transport_if<tlm::tlm_base_protocol_types>
  {
    typedef tlm::tlm_initiator_socket<32,
                                      tlm::tlm_base_protocol_types,
                                      MULTIPORT?0:1> base_type;
    typedef tlm::tlm_bw_transport_if<tlm::tlm_base_protocol_types> bw_if_type;

  public:
    initiator_socket():
      base_type(sc_core::sc_gen_unique_name(kind()))
    {
      init();
    }

    explicit initiator_socket(const char* name) :
      base_type(name)
    {
      init();
    }

    inline void init() {
      // Constant payload values
      m_payload.set_data_length(4);
      m_payload.set_byte_enable_length(4);
      m_payload.set_streaming_width(4);

      // we're not actually using the backward interface,
      // but we need to bind the sc_export of the socket to something.
      this->bind(*(static_cast<bw_if_type*>(this)));

      m_dmi = NULL;
    }

    ~initiator_socket() {}

    tlm::tlm_response_status read(const uint32_t& addr,
                                  uint32_t& data,
                                  uint32_t be_mask,
                                  sc_core::sc_time &delay,
                                  int port = 0)
    {
      // build the payload ...
      m_payload.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      m_payload.set_command(tlm::TLM_READ_COMMAND);
      m_payload.set_address(addr);
      m_payload.set_data_ptr(reinterpret_cast<unsigned char*>(&data));
      m_payload.set_byte_enable_ptr(reinterpret_cast<unsigned char*>(&be_mask));
      m_payload.set_dmi_allowed(false);

      // ... and send it.
      (*this)[port]->b_transport(m_payload, delay);
      
      return m_payload.get_response_status();
    }

    tlm::tlm_response_status write(const uint32_t& addr,
                                   uint32_t data,
                                   uint32_t be_mask,
                                   sc_core::sc_time &delay,
                                   int port = 0)
    {
      // build the payload ...
      m_payload.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      m_payload.set_command(tlm::TLM_WRITE_COMMAND);
      m_payload.set_address(addr);
      m_payload.set_data_ptr(reinterpret_cast<unsigned char*>(&data));
      m_payload.set_byte_enable_ptr(reinterpret_cast<unsigned char*>(&be_mask));
      m_payload.set_dmi_allowed(false);

      // ... and send it.
      (*this)[port]->b_transport(m_payload, delay);
      
      return m_payload.get_response_status();
    }

    virtual const char* kind() const {
      return "basic::initiator_socket";
    }

    // refer to last transaction
    inline bool is_dmi_allowed() const {
      return m_payload.is_dmi_allowed();
    }

    void invalidate_direct_mem_ptr(sc_dt::uint64, sc_dt::uint64)
    {
      if (m_dmi) m_dmi->clear();
    }

    tlm::tlm_sync_enum nb_transport_bw(tlm::tlm_generic_payload&,
                                       tlm::tlm_phase&,
                                       sc_core::sc_time&)
    {
      std::cerr << "nb_transport_bw not implemented" << std::endl;
      abort();
    }
    
    void set_dmi(HostMemory *dmi) {m_dmi = dmi;}

    tlm::tlm_response_status read_block_untimed(const uint32_t& addr,
                                                uint32_t* data,
                                                size_t tlm_size,
                                                size_t physical_size,
                                                sc_core::sc_time &delay,
                                                int port = 0)
    {
      // build the payload ...
      m_payload.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      m_payload.set_command(tlm::TLM_READ_COMMAND);
      m_payload.set_address(addr);
      m_payload.set_data_ptr(reinterpret_cast<unsigned char*>(data));
      m_payload.set_data_length(tlm_size);

      m_payload.set_byte_enable_length(0);
      m_payload.set_byte_enable_ptr(NULL);
      m_payload.set_dmi_allowed(false);

      untimed_extension ext(physical_size);
      m_payload.set_extension(&ext);

      // ... and send it.
      (*this)[port]->b_transport(m_payload, delay);

      // clean up the payload object
      m_payload.clear_extension(&ext);
      m_payload.set_data_length(4);      
      m_payload.set_byte_enable_length(4);

      return m_payload.get_response_status();
    }

    tlm::tlm_response_status write_block_untimed(const uint32_t& addr,
                                                 uint32_t *data,
                                                 size_t tlm_size,
                                                 size_t physical_size,
                                                 sc_core::sc_time &delay,
                                                 int port = 0)
    {
      // build the payload ...
      m_payload.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      m_payload.set_command(tlm::TLM_WRITE_COMMAND);
      m_payload.set_address(addr);
      m_payload.set_data_ptr(reinterpret_cast<unsigned char*>(data));
      m_payload.set_data_length(tlm_size);

      m_payload.set_byte_enable_length(0);
      m_payload.set_byte_enable_ptr(NULL);
      m_payload.set_dmi_allowed(false);

      untimed_extension ext(physical_size);
      m_payload.set_extension(&ext);

      // ... and send it.
      (*this)[port]->b_transport(m_payload, delay);
      
      // clean up the payload object
      m_payload.clear_extension(&ext);
      m_payload.set_data_length(4);      
      m_payload.set_byte_enable_length(4);

      return m_payload.get_response_status();
    }

  private:
    tlm::tlm_generic_payload m_payload;
    HostMemory *m_dmi;
  };

} // namespace basic

#endif // _INITIATOR_SOCKET_HPP_
