/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#ifndef _TARGET_SOCKET_HPP_
#define _TARGET_SOCKET_HPP_

#include "basic.hpp"
#include "trace.hpp"

namespace basic {

  typedef tlm::tlm_target_socket<32, tlm::tlm_base_protocol_types> compatible_socket;

  template <typename MODULE, bool MULTIPORT = false>
  class target_socket :
    public tlm::tlm_target_socket<32,
                                  tlm::tlm_base_protocol_types,
                                  MULTIPORT?0:1>,
    public tlm::tlm_fw_transport_if<tlm::tlm_base_protocol_types>
  {
    typedef tlm::tlm_target_socket<32,
                                   tlm::tlm_base_protocol_types,
                                   MULTIPORT?0:1> base_type;
    typedef tlm::tlm_fw_transport_if<tlm::tlm_base_protocol_types> fw_if_type;
  
  public:
    target_socket():
      base_type(sc_core::sc_gen_unique_name(kind()))
    {
      init();
    }

    explicit target_socket(const char* name):
      base_type(name)
    {
      init();
    }

    virtual const char* kind() const {
      return "basic::target_socket";
    }

    const char* name() const {
      return static_cast<const sc_core::sc_object*>(this)->name();
    }

    bool get_direct_mem_ptr(tlm::tlm_generic_payload&,
                            tlm::tlm_dmi&)
    {
      std::cerr << "get_direct_mem_ptr not implemented" << std::endl;
      abort();
    }

    unsigned int transport_dbg(tlm::tlm_generic_payload&) {
      std::cerr << "transport_dbg not implemented" << std::endl;
      abort();
    }

    tlm::tlm_sync_enum nb_transport_fw(tlm::tlm_generic_payload&,
                                       tlm::tlm_phase&,
                                       sc_core::sc_time&)
    {
      std::cerr << "nb_transport_fw not implemented" << std::endl;
      abort();
    }

  private:
    void b_transport(tlm::tlm_generic_payload& trans,
                     sc_core::sc_time&)
    {
      if (trans.get_data_length()!=4) {
        WARNING("Unsupported data length: "
                << trans.get_data_length());
        trans.set_response_status(tlm::TLM_BURST_ERROR_RESPONSE);
        return;
      }
      if (trans.get_byte_enable_ptr()) {
        if (trans.get_byte_enable_length()!=4 ||
            *reinterpret_cast<uint32_t*>(trans.get_byte_enable_ptr())
            != ~0u) {
          WARNING("Byte enable not supported by target");
          trans.set_response_status(tlm::TLM_BYTE_ENABLE_ERROR_RESPONSE);
          return;
        }
      }

      const uint32_t addr = static_cast<uint32_t>(trans.get_address());
      uint32_t& data =
        *(reinterpret_cast<uint32_t*>(trans.get_data_ptr()));

      switch(trans.get_command()) {
      case tlm::TLM_READ_COMMAND:
        trans.set_response_status(m_mod->read(addr, data));
        break;
      case tlm::TLM_WRITE_COMMAND:
        trans.set_response_status(m_mod->write(addr, data));
        break;
      case tlm::TLM_IGNORE_COMMAND:
        break;
      default:
        trans.set_response_status(tlm::TLM_COMMAND_ERROR_RESPONSE);
      }
    }

    void init() {
      // we'll receive transactions ourselves ...
      this->bind(*(static_cast<fw_if_type*>(this)));
      // ... but we'll need to call read/write in the parent module.
      m_mod = dynamic_cast<MODULE*>(this->get_parent_object());
      if(!m_mod) {
        std::cerr << this->name() << ": no parent" << std::endl;
        abort();
      }
    }
    MODULE* m_mod;
  };

} // namespace basic

#endif // _TARGET_SOCKET_HPP_
