//
// SimSoC Initial software, Copyright © INRIA 2007, 2008, 2009, 2010, 2011
// LGPL license version 3
//

#include "elf_loader.hpp"
#include "soclib_endian.hpp"
#include "trace.hpp"
#include "processor_if.hpp"

namespace {
  inline uint32_t bswap_32(uint32_t n) {
    return ::soclib::endian::uint32_swap(n);
  }
  inline uint16_t bswap_16(uint16_t n) {
    return ::soclib::endian::uint16_swap(n);
  }

  const char *name() {return "ELF_Loader";}
}

using namespace std;
using namespace sc_core;

namespace simsoc {

  /******************************************************************************/
  struct Elf32_SectionHeader: public Elf32_Shdr {

    const char *m_name;

    void unencode() {
      sh_name = bswap_32(sh_name);
      sh_type = bswap_32(sh_type);
      sh_flags = bswap_32(sh_flags);
      sh_addr = bswap_32(sh_addr);
      sh_offset = bswap_32(sh_offset);
      sh_size = bswap_32(sh_size);
      sh_link = bswap_32(sh_link);
      sh_info = bswap_32(sh_info);
      sh_addralign = bswap_32(sh_addralign);
      sh_entsize = bswap_32(sh_entsize);
    }

    size_t start() {
      return sh_addr;
    }

    size_t size() {
      return sh_size;
    }

    size_t file_offset() {
      return sh_offset;
    }

    const char *str_type(Elf32_Half machine) {
      switch (sh_type) {
      case SHT_NULL: return "NULL";
      case SHT_PROGBITS: return "PROGBITS";
      case SHT_SYMTAB: return "SYMTAB";
      case SHT_STRTAB: return "STRTAB";
      case SHT_NOBITS: return "NOBITS";
      default:
        INFO(2, "sh_type = " <<hex <<sh_type <<dec);
        return "other";
      }
    }

    bool is_load() {
      return (sh_flags&SHF_ALLOC) && sh_type==SHT_PROGBITS;
    }

    bool is_alloc(){
      return (sh_flags&SHF_ALLOC);
    }
  };

  /******************************************************************************/
  Elf32_Header::Elf32_Header() {
    strings = NULL;
    start_brk = 0;
  }

  Elf32_Header::~Elf32_Header() {
    for (size_t i = 0; i<sections.size(); ++i)
      delete sections[i];
    delete[] strings;
  }

  bool Elf32_Header::is_elf() {
    return
      e_ident[EI_MAG0]==ELFMAG0 &&
      e_ident[EI_MAG1]==ELFMAG1 &&
      e_ident[EI_MAG2]==ELFMAG2 &&
      e_ident[EI_MAG3]==ELFMAG3;
  }

  bool Elf32_Header::is_elf32() {
    return e_ident[EI_CLASS]==ELFCLASS32;
  }

  bool Elf32_Header::is_big_endian() {
    return e_ident[EI_DATA]==ELFDATA2MSB;
  }

  void Elf32_Header::unencode() {
    if (!is_big_endian())
      return;
    e_type = bswap_16(e_type);
    e_machine = bswap_16(e_machine);
    e_version = bswap_32(e_version);
    e_entry = bswap_32(e_entry);
    e_phoff = bswap_32(e_phoff);
    e_shoff = bswap_32(e_shoff);
    e_flags = bswap_32(e_flags);
    e_ehsize = bswap_16(e_ehsize);
    e_phentsize = bswap_16(e_phentsize);
    e_phnum = bswap_16(e_phnum);
    e_shentsize = bswap_16(e_shentsize);
    e_shnum = bswap_16(e_shnum);
    e_shstrndx = bswap_16(e_shstrndx);
  }

  bool Elf32_Header::is_exec() {
    return e_type==ET_EXEC;
  }

  uint32_t Elf32_Header::get_start_brk(){
    for (size_t i = 0; i<sections.size(); ++i) {
      if (sections[i]->is_alloc()) {
        uint32_t mem_start = sections[i]->start();
        uint32_t mem_size = sections[i]->size();
        uint32_t mem_end = mem_start+mem_size;
        if(start_brk < mem_end)
          start_brk = mem_end;
      }
    }
    return start_brk;
  }

  void Elf32_Header::load_sections(ifstream &ifs) {
    if (e_shentsize!=sizeof(Elf32_Shdr))
      ERROR("should be unreachable");
    ifs.seekg(e_shoff);
    if (!ifs)
      ERROR("should be unreachable");
    for (size_t i = 0; i<e_shnum; ++i) {
      sections.push_back(new Elf32_SectionHeader());
      ifs.read((char*) sections[i], sizeof(Elf32_Shdr));
      if (!ifs)
        ERROR("should be unreachable");
      if (is_big_endian())
        sections[i]->unencode();
    }
    if (e_shstrndx>=sections.size())
      ERROR("should be unreachable");
    strings = new char[sections[e_shstrndx]->size()];
    ifs.seekg(sections[e_shstrndx]->file_offset());
    if (!ifs)
      ERROR("should be unreachable");
    ifs.read(strings,sections[e_shstrndx]->size());
    if (!ifs)
      ERROR("should be unreachable");
    for (size_t i = 0; i<e_shnum; ++i) {
      sections[i]->m_name = strings+sections[i]->sh_name;
      INFO(2, "section " <<i <<": " <<sections[i]->m_name
           <<'\t' <<sections[i]->str_type(e_machine)
           <<(sections[i]->is_load()? " (load)": ""));
    }
  }

  /******************************************************************************/
  ElfFile::ElfFile(const char *elf_file):
    file_name(elf_file),
    ifs(file_name,ios::in|ios::binary),
    loaded(false),
    start_brk(0)
  {
    if (!ifs) {
      USER_ERROR("failed to open file \"" <<file_name <<'"');
    }
    ifs.read((char*) &header, sizeof(Elf32_Ehdr));
    if (!ifs)
      ERROR("should be unreachable");
    if (!header.is_elf()) {
      USER_ERROR('"' <<file_name <<"\" is not an ELF file");
    }
    if (!header.is_elf32()) {
      TODO("64-bit ELF file");
    }
    header.unencode();
    if (!header.is_exec()) {
      USER_ERROR('"' << file_name
                 << "\" is not an executable ELF file");
    }
    header.load_sections(ifs);

  }

  ElfFile::~ElfFile() {
    ifs.close();
    INFO(3, "ElfFile destructor called");
  }

  // Architecture ElfFile::get_architecture() {
  //   switch (header.e_machine) {
  //   case EM_MIPS:
  //     return MIPS;
  //   case EM_ARM:
  //     return ARM;
  //   case EM_PPC:
  //     return POWERPC;
  //   default:
  //     info() <<"e_machine = " <<header.e_machine <<endl;
  //     return LAST_ARCH;
  //   }
  // }

  bool ElfFile::is_big_endian() {
    return header.is_big_endian();
  }

  uint32_t ElfFile::get_start_address() {
    return header.e_entry;
  }

  void ElfFile::load_sections(char *dst, const size_t max) {
    for (size_t i = 0; i<header.sections.size(); ++i) {
      uint32_t mem_start = header.sections[i]->start();
      uint32_t mem_size = header.sections[i]->size();
      uint32_t mem_end = mem_start+mem_size;
      if (header.sections[i]->is_load()) {
        ifs.seekg(header.sections[i]->file_offset());
        if (!ifs)
          ERROR("should be unreachable");
        INFO(2, hex <<"write from " <<mem_start <<" to "
             <<mem_end <<dec);
        char * tmp = new char[mem_size];
        ifs.read(tmp, mem_size);
        if (!ifs)
          ERROR("should be unreachable");
        for (uint32_t j = 0; j<mem_size; ++j) {
          uint64_t addr;
          addr = mem_start+j;
          assert(addr<max);
          dst[addr] = tmp[j];
        }
        delete[] tmp;
      }
      if(header.sections[i]->is_alloc()){
        if(start_brk < mem_end)
          start_brk = mem_end;
      }
    }
    loaded=true;
  }

  uint32_t ElfFile::get_elf_brk(){
    if(loaded)
      return start_brk;
    else{
      //header sections already been loaded in ElfFile Constructor
      start_brk=header.get_start_brk();
      return start_brk;
    }
  }


  /*****************************************************************************/
  void ElfLoader::load_sections(ElfFile &elf_file, int socket_id) {
    tlm::tlm_generic_payload trans;
    trans.set_streaming_width(1);
    trans.set_command(tlm::TLM_WRITE_COMMAND);
    if (elf_file.loaded)
      return;
    for (size_t i = 0; i<elf_file.header.sections.size(); ++i) {
      uint32_t mem_start = elf_file.header.sections[i]->start();
      uint32_t mem_size = elf_file.header.sections[i]->size();
      uint32_t mem_end = mem_start+mem_size;
      if (elf_file.header.sections[i]->is_load()) {
        elf_file.ifs.seekg(elf_file.header.sections[i]->file_offset());
        if (!elf_file.ifs)
          ERROR("should be unreachable");
        INFO(2, hex <<"write from " <<mem_start <<" to "
             <<mem_end <<dec);
        char * tmp = new char[mem_size];
        elf_file.ifs.read(tmp, mem_size);
        if (!elf_file.ifs)
          ERROR("should be unreachable");
        trans.set_data_length(mem_size);
        trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
        trans.set_address(mem_start);
        trans.set_data_ptr(reinterpret_cast<uint8_t*>(tmp));
        const unsigned int n = initiator_sockets[socket_id]->transport_dbg(trans);
        if (n!=mem_size ||
            trans.get_response_status()!=tlm::TLM_OK_RESPONSE) {
          ERROR("Only " <<n <<" bytes copied: "
                <<trans.get_response_string());
        }
        delete[] tmp;
      }
      if (elf_file.header.sections[i]->is_alloc()){
        if (elf_file.start_brk < mem_end)
          elf_file.start_brk = mem_end;
      }
    }
    elf_file.loaded = true;
  }

  ElfLoader::ElfLoader(sc_core::sc_module_name):
    jobs()
  {}

  void ElfLoader::set_file(ProcessorIF *proc,
                           const string &elf_file) {
    assert(proc && "invalid parameters");
    int sid = jobs.size();
    initiator_sockets.bind(proc->get_debug_target());
    if (elf_file!="none")
      jobs.push_back(Job(proc,elf_file,sid));
  }

  void ElfLoader::end_of_elaboration() {
    for (size_t i=0, ei=jobs.size(); i!=ei; ++i) {
      ProcessorIF *proc = jobs[i].proc;
      ElfFile elf_file(jobs[i].filename.c_str());
      if (elf_file.is_big_endian())
        proc->set_endianness(ProcessorIF::BIG_ENDIAN_32);
      else
        proc->set_endianness(ProcessorIF::LITTLE_ENDIAN_32);
      load_sections(elf_file,jobs[i].socket_id);
      proc->set_pc(elf_file.get_start_address());
    }
  }

} // namespace simsoc
