/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "hwgol.hpp"
#include "hwgol_regmap.h"
#include "vga_regmap.h"

using namespace std;
using namespace sc_core;

namespace {

  inline int TEST_BIT(uint32_t w, unsigned i) {
    return 1&(w>>i);
  }
  
  inline void SET_BIT(uint32_t &w, unsigned i) {
    w |= (1<<i);
  }
  
  inline void CLEAR_BIT(uint32_t &w, unsigned i) {
    w &=~ (1<<i);
  }

  const sc_time time_per_image(150, SC_MS);

  const double COMPUTING = 0.9;
  const double WAITING = 0.1;

  const size_t SW_VGA_HEIGHT = VGA_HEIGHT /2;
  const size_t SW_VGA_WIDTH = VGA_WIDTH /2;

  int get_pixel(const uint32_t *buffer, int x, int y) {
    const uint32_t offset = (y * (VGA_LINE / 32)) + x / 32;
    const int bit = 31 - (x % 32);
    return TEST_BIT(buffer[offset], bit)!=0;
  }

  void set_pixel(uint32_t *buffer, int x, int y, int value) {
    const uint32_t offset = (y * (VGA_LINE / 32)) + x / 32;
    int bit = 31 - (x % 32);
    
    uint32_t &data = buffer[offset];
    if (value) {
      SET_BIT(data, bit);
    } else {
      CLEAR_BIT(data, bit);
    }
  }

  unsigned neighbours(const uint32_t *buffer, unsigned x, unsigned y) {
    unsigned n = 0;
    if (x > 0) {
      if(y > 0) {
        n += get_pixel(buffer, x - 1, y - 1);
      }
      n += get_pixel(buffer, x - 1, y);
      if (y < SW_VGA_HEIGHT-1) {
        n += get_pixel(buffer, x - 1, y + 1);
      }
    }
    if (y > 0) {
      n += get_pixel(buffer, x, y - 1);
    }
    if (y < SW_VGA_HEIGHT-1) {
      n += get_pixel(buffer, x, y + 1);
    }
    if (x < SW_VGA_WIDTH-1) {
      if (y > 0) {
        n += get_pixel(buffer, x + 1, y - 1);
      }
      n += get_pixel(buffer, x + 1, y);
      if (y < SW_VGA_HEIGHT-1) {
        n += get_pixel(buffer, x + 1, y + 1);
      }
    }
    return n;
  }

  void game_of_life(const uint32_t *read_buffer, uint32_t* write_buffer)
  {
    for (unsigned x = 0; x < SW_VGA_WIDTH; ++x) {
      for (unsigned y = 0; y < SW_VGA_HEIGHT; ++y) {
        const unsigned n = neighbours(read_buffer, x, y);
        /* if is alive */
        if (get_pixel(read_buffer, x, y)) {
          /* underpopulation or overcrowding: die */
          if ((n < 2) || (n > 3)) {
            set_pixel(write_buffer, x, y, 0);
          } else {
            // set_pixel(write_buffer, x, y, 1);
          }
        } else {
          /* perfect conditions: become alive */
          if (n == 3) {
            set_pixel(write_buffer, x, y, 1);
          } else {
            // set_pixel(write_buffer, x, y, 0);
          }
        }
      }
    }
  }
}

HwGoL::HwGoL(sc_core::sc_module_name):
  PwtModule(this),
  m_running(false),
  m_data_src(0), m_data_dst(0)
{
  INFO(2, "in constructor");
  SC_THREAD(compute);
  irq.initialize(true);
}

HwGoL::~HwGoL() {}

void HwGoL::compute() {
  set_activity(WAITING);
  while (true) {
    while (!m_running) 
      wait(start_event);
    INFO(1, "start computation");
    set_activity(COMPUTING);
    m_running = true;
    irq.write(false);
    const size_t nb_word = VGA_LINE/32 * (SW_VGA_HEIGHT);
    const size_t nb_byte = nb_word * 4;
    uint32_t *read_buffer = new uint32_t[nb_word];
    std::fill(read_buffer, read_buffer+nb_word, 0);
    sc_time zero = SC_ZERO_TIME;

    tlm::tlm_response_status status;
    status = initiator.read_block_untimed(m_data_src, read_buffer, nb_byte, 9*nb_byte, zero);
    assert(status==tlm::TLM_OK_RESPONSE && "fail to read image");

    uint32_t *write_buffer = new uint32_t[nb_word];
    std::copy(read_buffer, read_buffer+nb_word, write_buffer);
    game_of_life(read_buffer, write_buffer);

    status = initiator.write_block_untimed(m_data_dst, write_buffer, nb_byte, 1*nb_byte, zero);
    assert(status==tlm::TLM_OK_RESPONSE && "fail to write image");
    assert(zero==SC_ZERO_TIME && "untimed transactions should not change the delay");

    delete[] read_buffer; delete[] write_buffer;

    DelayedActivityManager::commit(time_per_image);
    wait(time_per_image);

    set_activity(WAITING);
    m_running = false;
    INFO(1, "raise irq");
    irq.write(true); trace_irq();
  }
}

tlm::tlm_response_status HwGoL::read(uint32_t addr, uint32_t& data) {
  switch (addr) {
  case HWGOL_START_REG:
    WARNING("write-only register");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
    break;
  case HWGOL_STATUS_REG:
    data = m_running;
    break;
  case HWGOL_DATA_SRC_REG:
    data = m_data_src;
    break;
  case HWGOL_DATA_DST_REG:
    data = m_data_dst;
    break;
  default:
    WARNING("register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}
  
tlm::tlm_response_status HwGoL::write(uint32_t addr, uint32_t data) {
  switch (addr) {
  case HWGOL_START_REG:
    assert(m_running==false && "invalid ESW command");
    INFO(1, "receive START command");
    m_running = true;
    start_event.notify();
    break;
  case HWGOL_STATUS_REG:
    WARNING("read-only register");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
    break;
  case HWGOL_DATA_SRC_REG:
    assert(m_running==false && "invalid ESW command");
    m_data_src = data;
    break;
  case HWGOL_DATA_DST_REG:
    assert(m_running==false && "invalid ESW command");
    m_data_dst = data;
    break;
  default:
    WARNING("register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}
