/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "intc.hpp"
#include "intc_regmap.h"
#include "trace.hpp"

using namespace std;
using namespace sc_core;

namespace {
  ParameterInt
  intc_period("intc-period",
              "Clock period of the interrupt controller (in ns)",
              20);
}

Intc::Intc(sc_core::sc_module_name, unsigned nb_in):
  PwtModule(this),
  in_ports(new sc_in<bool>*[nb_in]),
  out("out"),
  m_nb_in(nb_in),
  event_catchers(new EventCatcher*[nb_in]),
  m_enabled_it(0),
  m_active_it(0),
  irq_to_send_notified(false)
{
  SC_THREAD(send_irq);
  sensitive << irq_to_send;
  out.initialize(false);

  for (unsigned i = 0; i<nb_in; ++i) {
    ostringstream os;
    os << "in_" << i;
    in_ports[i] = new sc_in<bool>(os.str().c_str());
    os.str("");
    os << "event_catcher_" << i;
    event_catchers[i] = new EventCatcher(os.str().c_str(), this, i);
  }
}

Intc::~Intc() {
  for (unsigned i = 0; i<m_nb_in; ++i) {
    delete in_ports[i];
    delete event_catchers[i];
  }
  delete[] in_ports;
  delete[] event_catchers;
}

Intc::EventCatcher::EventCatcher(sc_module_name,
                                 Intc *parent,
                                 unsigned id):
  m_parent(*parent), m_id(id)
{
  SC_METHOD(process_in_irq);
  sensitive << m_parent.in_ports[id]->pos();
  dont_initialize();
}

void Intc::EventCatcher::process_in_irq() {
  m_parent.process_in_irq(m_id);
}

void Intc::process_in_irq(unsigned id) {
  INFO(2, "Process_in_irq (" << id << hex << ", " << m_enabled_it
       << " & " << m_active_it << " = "
       << (m_enabled_it & m_active_it) << ")" << dec);
  const uint32_t mask = 1<<id;
  m_active_it |= mask;
  send_irq_maybe(mask);
}

void Intc::send_irq_maybe(uint32_t mask) {
  if (m_enabled_it & m_active_it & mask) {
    irq_to_send_notified = true;
    irq_to_send.notify();
    INFO(2, "Send IRQ maybe (" << hex << m_enabled_it
         << " & " << m_active_it << " = "
         << (m_enabled_it & m_active_it)
         << "," << mask << ")" << dec);
  }
}

void Intc::send_irq() {
  while (true) {
    while (!irq_to_send_notified) {
      wait();
    }
    irq_to_send_notified = false;
    INFO(1, "Send IRQ (" << hex << m_enabled_it
         << " & " << m_active_it << " = "
         << (m_enabled_it & m_active_it) << ")" << dec);
    out.write(true); trace_irq();
    wait(intc_period.get(), SC_NS);
    out.write(false);
    wait(intc_period.get(), SC_NS);
  }
}

tlm::tlm_response_status
Intc::read(uint32_t addr, uint32_t& data)
{
  switch (addr) {
  case INTC_ISR:
    data = m_active_it;
    break;
  case INTC_IER:
    data = m_enabled_it;
    break;
  case INTC_IAR:
    WARNING("read to a write-only register");
    return tlm::TLM_COMMAND_ERROR_RESPONSE;
    break;
  default:
    WARNING("register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}

tlm::tlm_response_status
Intc::write(uint32_t addr, uint32_t data)
{
  switch (addr) {
  case INTC_ISR:
    WARNING("write to a read-only register");
    return tlm::TLM_COMMAND_ERROR_RESPONSE;
    break;
  case INTC_IER: {
    const uint32_t old_data = data;
    INFO(2, "enable interrupt: " << hex << data << dec);
    m_enabled_it = data;
    send_irq_maybe(data & ~(old_data));
    break;
  }
  case INTC_IAR:
    INFO(1, "Acknowledging IT (" << hex << m_active_it << ")"
         << dec);
    m_active_it &= ~data;
    send_irq_maybe(data);
    break;
  case INTC_SIR:
    // Software Interrupt Register. Doesn't exist
    // in real hardware unfortunately.
    INFO(2, "Software interrupt: " << hex << data << dec);
    m_active_it |= data;
    send_irq_maybe(data);
    break;
  default:
    WARNING("register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}
