//
// SimSoC Initial software, Copyright © INRIA 2007, 2008, 2009, 2010, 2011
// LGPL license version 3
//

// Adapted for LIBTLMPWT (Verimag/CNRS, 2012, 2013), by Claude Helmstetter

#include "simsoc_bus.hpp"
#include "trace.hpp"
#include <algorithm>
#include "ecounters_ext.hpp"
#include "untimed_ext.hpp"

using namespace std;
using namespace sc_core;
using namespace tlm;

namespace simsoc {

  SimSoCBus::SimSoCBus(sc_module_name name):
    PwtTargetModule(this, .8, .8, .8, .8),
    target_sockets("target_sockets"),
    initiator_sockets("initiator_sockets"),
    intervals()
  {
    target_sockets.register_b_transport(this, &SimSoCBus::b_transport);
    target_sockets.register_nb_transport_fw(this, &SimSoCBus::nb_transport_fw);
    target_sockets.register_get_direct_mem_ptr(this, &SimSoCBus::get_direct_mem_ptr);
    target_sockets.register_transport_dbg(this, &SimSoCBus::transport_dbg);
    initiator_sockets.register_nb_transport_bw(this, &SimSoCBus::nb_transport_bw);
    initiator_sockets.register_invalidate_direct_mem_ptr(this, &SimSoCBus::invalidate_dm_ptr);
  }

  void SimSoCBus::before_end_of_elaboration() {
    // sort intervals
    std::sort(intervals.begin(),intervals.end());
    // initialize "previous" iterator
    previous = intervals.begin();
    // check intervals for overlaps
    for (size_t i = 1, ei = intervals.size(); i!=ei; ++i)
      if (intervals[i].last>=intervals[i-1].addr) {
        print_mmap();
        ERROR("at least two intervals overlap in memory map.");
      }
    if (trace_level.get()>=1)
      print_mmap();
  }

  void SimSoCBus::bind_target(base_target_socket_type &port,
                              uint64_t addr,
                              uint64_t size)
  {
    assert(addr<=addr+size-1);
    size_t n = intervals.size();
    assert((size_t)initiator_sockets.size()==n);
    initiator_sockets(port);
    intervals.push_back(Interval(n,addr,size));
  }

  void SimSoCBus::b_transport(int id,
                              tlm::tlm_generic_payload& pl,
                              sc_time& delay)
  {
    assert(pl.get_response_status()==TLM_INCOMPLETE_RESPONSE);
    uint64_t addr = pl.get_address();
    size_t index;
    if (resolve(addr,index)) {
      pl.set_address(addr);
      untimed_extension *ext;
      pl.get_extension(ext);
      if (ext) {
        add_delayed_side_effects(ext->get_size(), pl.get_command(), delay);
      } else {   
        apply_side_effects(pl.get_data_length()/4, pl.get_command(), delay);
      }
      initiator_sockets[index]->b_transport(pl, delay);
    } else {
      pl.set_response_status(TLM_ADDRESS_ERROR_RESPONSE);
    }
    INFO(3, "b_transport: " << pl << " @ T + " << delay);
  }

  tlm::tlm_sync_enum
  SimSoCBus::nb_transport_fw(int id,
                             tlm::tlm_generic_payload& pl,
                             tlm::tlm_phase& phase,
                             sc_time& delay)
  {
    uint64_t addr = pl.get_address();
    size_t index;
    if (resolve(addr,index)) {
      route_extension* ext = 0;
      if (phase==tlm::BEGIN_REQ) {
        ext = new route_extension;
        ext->id = id;
        accessor(pl).set_extension(ext);
      }
      pl.set_address(addr);
      tlm::tlm_sync_enum status;
      untimed_extension *ut_ext;
      pl.get_extension(ut_ext);
      if (ut_ext) {
        add_delayed_side_effects(ut_ext->get_size(), pl.get_command(), delay);
      } else {   
        apply_side_effects(pl.get_data_length()/4, pl.get_command(), delay);
      }
      status =
        initiator_sockets[index]->nb_transport_fw(pl, phase, delay);
      if (status==tlm::TLM_COMPLETED) {
        accessor(pl).clear_extension(ext);
        delete ext;
      }
      return status;
    } else {
      pl.set_response_status(TLM_ADDRESS_ERROR_RESPONSE);
      return TLM_COMPLETED;
    }
  }

  tlm::tlm_sync_enum
  SimSoCBus::nb_transport_bw(int id, tlm::tlm_generic_payload& pl,
                             tlm::tlm_phase& phase, sc_time& delay)
  {
    route_extension* ext = 0;
    accessor(pl).get_extension(ext);
    assert(ext);
    tlm::tlm_sync_enum status;
    status =
      target_sockets[ext->id]->nb_transport_bw(pl, phase, delay);
    if (status==tlm::TLM_COMPLETED) {
      accessor(pl).clear_extension(ext);
      delete ext;
    }
    return status;
  }

  unsigned int SimSoCBus::transport_dbg(int id,
                                        tlm::tlm_generic_payload& pl)
  {
    assert(pl.get_response_status()==TLM_INCOMPLETE_RESPONSE);
    uint64_t addr = pl.get_address();
    size_t index;
    if (resolve(addr,index)) {
      pl.set_address(addr);
      const unsigned int reply =
        initiator_sockets[index]->transport_dbg(pl);
      INFO(3, "transport_dbg[" << reply << "]: " << pl);
      return reply;
    } else {
      pl.set_response_status(TLM_ADDRESS_ERROR_RESPONSE);
      INFO(3, "transport_dbg[0]: " << pl);
      return 0;
    }
  }

  bool SimSoCBus::get_direct_mem_ptr(int id,
                                     tlm::tlm_generic_payload& pl,
                                     tlm::tlm_dmi& dmi_data)
  {
    assert(pl.get_response_status()==TLM_INCOMPLETE_RESPONSE);
    uint64_t addr = pl.get_address();
    const uint64_t orig_addr = addr;
    uint64_t base = 0;
    size_t index;
    if (resolve(addr,index,base)) {
      pl.set_address(addr);
      bool status =
        initiator_sockets[index]->get_direct_mem_ptr(pl, dmi_data);
      uint64_t dmi_start = base + dmi_data.get_start_address();
      uint64_t dmi_end = base + dmi_data.get_end_address();
      dmi_data.set_start_address(dmi_start);
      dmi_data.set_end_address(dmi_end);
      pl.set_address(orig_addr);
      side_effect_extension *ext;
      pl.get_extension(ext);
      if (ext) ext->add_side_effect(this);
      return status;
    } else {
      pl.set_response_status(TLM_ADDRESS_ERROR_RESPONSE);
      return false;
    }
  }

  void SimSoCBus::invalidate_dm_ptr(int id,
                                    sc_dt::uint64 start_range,
                                    sc_dt::uint64 end_range)
  {
    size_t i = 0, ei = intervals.size();
    while (i!=ei && id!=(int)intervals[i].index) ++i;
    assert(i!=ei && id==(int)intervals[i].index && "invalid id");
    assert(start_range<intervals[i].size()
           && "start address outside target range");
    const uint64_t start = intervals[i].addr + start_range;
    if (end_range>intervals[i].size()) {
      WARNING("invalidate_direct_mem_ptr: "
              <<"end_range is outside target size.");
      end_range = intervals[i].size();
    }
    const uint64_t end = intervals[i].addr + end_range;
    for (unsigned int i = 0; i<target_sockets.size(); ++i)
      target_sockets[i]->invalidate_direct_mem_ptr(start, end);
  }

  void SimSoCBus::print_mmap(ostream &os) {
    for (vector<SimSoCBus::Interval>::iterator i = intervals.begin(),
           ei = intervals.end(); i!=ei; ++i) {
      os <<name() << ": interval ["
         <<hex <<i->addr <<',' <<i->last+1
         <<") mapped to port #" <<dec <<i->index;
      const sc_object *obj =
        dynamic_cast<const sc_object*>(initiator_sockets[i->index]);
      if (obj)
        os <<" \"" << obj->name() <<"\"";
      os <<endl;
    }
  }

  // return true if a target is found at this address
  bool SimSoCBus::resolve(uint64_t &addr, size_t &index)
  {
    if (previous->addr<=addr && addr<=previous->last) {
      addr -= previous->addr;
      index = previous->index;
      return true;
    }
    std::vector<SimSoCBus::Interval>::iterator i =
      lower_bound(intervals.begin(),
                  intervals.end(),
                  Interval(addr));
    if (i==intervals.end() || i->last<addr) {
      return false;
    } else {
      previous = i;
      addr -= i->addr;
      index = i->index;
      return true;
    }
  }

  // return true if a target is found at this address
  bool SimSoCBus::resolve(uint64_t &addr, size_t &index,
                          uint64_t &base) {
    std::vector<SimSoCBus::Interval>::iterator i =
      lower_bound(intervals.begin(),
                  intervals.end(),
                  Interval(addr));
    if (i==intervals.end() || i->last<addr) {
      return false;
    } else {
      addr -= i->addr;
      index = i->index;
      base = i->addr;
      return true;
    }
  }

  void SimSoCBus::set_frequency_hook(unsigned) {
    for (unsigned int i = 0; i<target_sockets.size(); ++i)
      target_sockets[i]->invalidate_direct_mem_ptr(0, ~0llu);
  }

} // namespace simsoc
