/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "temp_sensor.hpp"
#include "temp_sensor_regmap.h"
#include "trace.hpp"

using namespace sc_core;
using namespace std;;

// we store temperature with two decimal
// places; i.e., xx.dd as an integer value
// (mutiply by 100)

TempSensor::TempSensor(sc_core::sc_module_name name):
  PwtModule(this),
  m_current_temp(INITIAL_TEMPERATURE * 100),
  m_temp_high(100 * 100),
  m_temp_low(0),
  m_interrupt(0)
{
  SC_THREAD(send_interrupt);
  irq.initialize(false);
}

// Signals an interrupt each time the solver
// states that a threshold is reached.
void TempSensor::send_interrupt() {
  while (true) {
    wait(m_interrupt_event);
    if (m_interrupt!=1) {
      INFO(2, "SENDING INTERRUPT, T = " << m_current_temp << " /100 °C");
      m_interrupt = 1;
      irq.write(true); trace_irq();
      wait(sc_core::SC_ZERO_TIME);
      irq.write(false);
    }
  }
}

tlm::tlm_response_status
TempSensor::read(const uint32_t &addr, uint32_t& data)
{
  switch (addr) {
  case CURRENT_TEMP_OFFSET:
    data = m_current_temp;
    INFO(2, "Read Temperature returns " << data);
    break;
  case TEMP_HIGH_OFFSET:
    data = m_temp_high;
    break;
  case TEMP_LOW_OFFSET:
    data = m_temp_low;
    break;
  case TEMP_INT_OFFSET:
    data = m_interrupt;
    INFO(2, "Read Interrupt Register returns " << data);
    break;
  default:
    WARNING("register " << hex << addr << dec << " not implemented for READ access");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}

tlm::tlm_response_status
TempSensor::write(const uint32_t &addr, const uint32_t &data)
{
  switch (addr) {
  case CURRENT_TEMP_OFFSET:
    WARNING("register CURRENT_TEMP is not accessible in write mode");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
    break;
  case TEMP_HIGH_OFFSET:
    m_temp_high = data;
    INFO(2, "HIGH temp updated to " << m_temp_high << " /100 °C");
    break;
  case TEMP_LOW_OFFSET:
    m_temp_low = data;
    INFO(2, "LOW temp updated to " << m_temp_low << " /100 °C");
    break;
  case TEMP_INT_OFFSET:
    m_interrupt = data;
    if (data==0) {
      INFO(2, "interrupt cleared");
    } else {
      INFO(2, "Wrote to the interrupt register with a value different from 0");
    }
    break;
  default:
    WARNING("register not implemented for WRITE");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}

// This function is called by the simulation/temp-solver
// syncronization module to set the current temperature time
void TempSensor::set_temperature_hook(double temp)
{
  INFO(2, "Updating Temperature to " << temp << " °C");
  m_current_temp = temp * 100;

  if (m_interrupt==0 &&
      (m_current_temp < m_temp_low || m_temp_high < m_current_temp)) {
    m_interrupt_event.notify();
  }
}
