/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "top.hpp"
#include "memory_map.h"
#include "memory.hpp"
#include "uart.hpp"
#include "timer.hpp"
#include "simsoc_bus.hpp"
#include "trace.hpp"
#include "parameter.hpp"
#include "mb_wrapper.hpp"
#include "elf_loader.hpp"
#include "intc.hpp"
#include "intc_regmap.h"
#include "headless_vga.hpp"
#include "vga.hpp"
#include "hwgol.hpp"
#include "temp_sensor.hpp"
#include "power_ctrl.hpp"
#include "power_ctrl_regmap.h"
#include "atmi_wrapper.hpp"

using namespace std;
using namespace sc_core;
using namespace simsoc;

ParameterBool
screen("screen",
       "Enable the VGA screen");

ParameterString
elf_file("esw",
         "ELF file containing the microblaze embedded software",
         "none");

Top::Top(sc_module_name):
  PwtModule(this),
  atmi_wrapper(new AtmiWrapper("ATMI_WRAPPER", 10)),
  bus(new SimSoCBus("BUS")),
  sram(new Memory("SRAM", SRAM_SIZE)),
  inst_ram(new Memory("INST_RAM", INST_RAM_SIZE)),
  uart(new Uart("UART")),
  timer(new Timer("TIMER", sc_time(20, SC_NS))),
  mb_wrapper(new MBWrapper("MB_CPU", 0)),
  elf_loader(new ElfLoader("ELF_LOADER")),
  intc(new Intc("INTC", INTC_IN_COUNT)),
  vga(screen.get() ? new Vga("VGA") : new HeadlessVGA("VGA")),
  hwgol(new HwGoL("HWGOL")),
  temp_sensor(new TempSensor("TEMP_SENSOR")),
  power_ctrl(new PowerCtrl("POWER_CTRL", 2)),
  vga_sig("vga_signal"),
  timer_sig("timer_signal"),
  temp_sensor_sig("temp_sensor_signal"),
  hwgol_sig("hwgol_signal"),
  mb_irq_sig("mb_irq_signal")
{
  INFO(1, "in constructor");

  // Initiators -> Bus
  vga->initiator.bind(bus->target_sockets);
  hwgol->initiator.bind(bus->target_sockets);
  mb_wrapper->initiator.bind(bus->target_sockets);

  // Bus -> Targets
  bus->bind_target(sram->target,
                   SRAM_BASEADDR, SRAM_SIZE);
  bus->bind_target(inst_ram->target,
                   INST_RAM_BASEADDR, INST_RAM_SIZE);
  bus->bind_target(uart->target,
                   UART_BASEADDR, UART_SIZE);
  bus->bind_target(timer->target,
                   TIMER_BASEADDR, TIMER_SIZE);
  bus->bind_target(intc->target,
                   INTC_BASEADDR, INTC_SIZE);
  bus->bind_target(vga->target,
                   VGA_BASEADDR, VGA_SIZE);
  bus->bind_target(hwgol->target,
                   HWGOL_BASEADDR, HWGOL_SIZE);
  bus->bind_target(temp_sensor->target,
                   TEMP_SENSOR_BASEADDR, TEMP_SENSOR_SIZE);
  bus->bind_target(power_ctrl->target,
                   POWER_CTRL_BASEADDR, POWER_CTRL_SIZE);

  // Wires
  vga->irq.bind(vga_sig);
  timer->irq.bind(timer_sig);
  temp_sensor->irq.bind(temp_sensor_sig);
  hwgol->irq.bind(hwgol_sig);
  mb_wrapper->irq.bind(mb_irq_sig);

  intc->in_ports[VGA_IRQ]->bind(vga_sig);
  intc->in_ports[TIMER_IRQ]->bind(timer_sig);
  intc->in_ports[SENSOR_IRQ]->bind(temp_sensor_sig);
  intc->in_ports[HWGOL_IRQ]->bind(hwgol_sig);
  intc->out.bind(mb_irq_sig);
  
  // DVFS
  power_ctrl->set_module(PWCTRL_ID_TOP, this,
                         PWCTRL_VOLT_NORMAL, PWCTRL_FREQ_NORMAL);
  power_ctrl->set_module(PWCTRL_ID_MBn(0), mb_wrapper,
                         PWCTRL_VOLT_NORMAL, PWCTRL_FREQ_NORMAL);

  // ELF loader  
  elf_loader->set_file(mb_wrapper, elf_file.get());

  // ATMI: floorplan and traces
  atmi_wrapper->config_rectangle(sram, 0.5, 0, 0.9, 0.2);
  atmi_wrapper->config_rectangle(inst_ram, 0, 0, 0.4, 0.2);
  atmi_wrapper->config_rectangle(bus, 0.4, 0, 0.5, 0.9);
  atmi_wrapper->config_rectangle(power_ctrl, 0, 0.7, 0.4, 0.9);
  atmi_wrapper->config_rectangle(vga, 0, 0.35, 0.4, 0.55);
  atmi_wrapper->config_rectangle(temp_sensor, 0, 0.2, 0.4, 0.35);
  atmi_wrapper->config_rectangle(intc, 0, 0.55, 0.4, 0.7);
  atmi_wrapper->config_rectangle(mb_wrapper, 0.5, 0.2, 0.9, 0.6);
  atmi_wrapper->config_rectangle(timer, 0.5, 0.6, 0.9, 0.7);
  atmi_wrapper->config_rectangle(hwgol, 0.5, 0.7, 0.9, 0.9);
  vga->enable_trace();
  temp_sensor->enable_trace();
  mb_wrapper->enable_trace();
  intc->enable_irq_trace();
}

Top::~Top() {
  INFO(1, "in destructor");
  delete bus;
  delete sram;
  delete inst_ram;
  delete uart;
  delete timer;
  delete elf_loader;
  delete intc;
  delete vga;
  delete hwgol;
  delete temp_sensor;
  delete power_ctrl;
  delete atmi_wrapper;
  delete mb_wrapper;
}
