/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>

#include "memory_map.h"
#include "vga_regmap.h"
#include "timer_regmap.h"
/* #include "gpio_regmap.h" */
#include "intc_driver.h"
#include "bit.h"
#include "hal.h"

#define SW_VGA_WIDTH  (VGA_WIDTH /4)  // "/4" <= pour accélérer
#define SW_VGA_HEIGHT (VGA_HEIGHT /4) //         la simulation

// Parallelism
#ifndef NB_SLAVE
#define NB_SLAVE 0
#endif

#ifndef NB_SLICES
#define NB_SLICES (NB_SLAVE + 1)
#endif

#ifndef SLICE_WIDTH
/* #error "Cannot compute SLICE_WIDTH myself, sorry" */
#define SLICE_WIDTH (SW_VGA_WIDTH/NB_SLICES)
#endif

int irq_received = 0;
int should_redraw[NB_SLAVE];

int compute = 0;

int get_pixel(uint32_t base_addr, int x, int y) {
  uint32_t addr = base_addr +
    ALIGN( (x / CHAR_BIT) + (y * (VGA_LINE / CHAR_BIT)) );
  int bit = 31 - x % (sizeof(uint32_t) * CHAR_BIT);

  uint32_t data = read_mem(addr);
  return TEST_BIT(data, bit)!=0;
}

void set_pixel(uint32_t base_addr, int x, int y, int value) {
  uint32_t addr = base_addr +
    ALIGN( (x / CHAR_BIT) + (y * (VGA_LINE / CHAR_BIT)) );
  int bit = 31 - x % (sizeof(uint32_t) * CHAR_BIT);

  uint32_t data = read_mem(addr);
  if (value) {
    SET_BIT(&data, bit);
  } else {
    CLEAR_BIT(&data, bit);
  }
  write_mem(addr, data);
}

void clr_screen(uint32_t base_addr) {
  int i;
  int nbytes = (VGA_LINE * VGA_HEIGHT) / CHAR_BIT;
  for (i = 0; i < nbytes; i += sizeof(uint32_t)) {
    write_mem(base_addr + i, 0);
  }
}

void blank_screen(uint32_t base_addr) {
  int i;
  int nbytes = (VGA_LINE * VGA_HEIGHT) / CHAR_BIT;
  for(i = 0; i < nbytes; i += sizeof(uint32_t)) {
    write_mem(base_addr + i, 0xFFFFFFFF);
  }
}

int neighbours(uint32_t base_addr, int x, int y) {
  int n = 0;
  if (x > 0) {
    if(y > 0) {
      n += get_pixel(base_addr, x - 1, y - 1);
    }
    n += get_pixel(base_addr, x - 1, y);
    if (y < SW_VGA_HEIGHT-1) {
      n += get_pixel(base_addr, x - 1, y + 1);
    }
  }
  if (y > 0) {
    n += get_pixel(base_addr, x, y - 1);
  }
  if (y < SW_VGA_HEIGHT-1) {
    n += get_pixel(base_addr, x, y + 1);
  }
  if (x < SW_VGA_WIDTH-1) {
    if (y > 0) {
      n += get_pixel(base_addr, x + 1, y - 1);
    }
    n += get_pixel(base_addr, x + 1, y);
    if (y < SW_VGA_HEIGHT-1) {
      n += get_pixel(base_addr, x + 1, y + 1);
    }
  }
  return n;
}

int refresh;
uint32_t new_img_addr;
uint32_t old_img_addr;

void game_of_life(int part) {
  int x;
  int y;
  for (x = (part * SLICE_WIDTH);
       x < ((part + 1) * SLICE_WIDTH);
       ++x) {
    for (y = 0; y < SW_VGA_HEIGHT; ++y) {
      int n = neighbours(old_img_addr, x, y);
      /* if is alive */
      if (get_pixel(old_img_addr, x, y)) {
        /* underpopulation or overcrowding: die */
        if ((n < 2) || (n > 3)) {
          set_pixel(new_img_addr, x, y, 0);
        } else {
          set_pixel(new_img_addr, x, y, 1);
        }
      } else {
        /* perfect conditions: become alive */
        if (n == 3) {
          set_pixel(new_img_addr, x, y, 1);
        } else {
          set_pixel(new_img_addr, x, y, 0);
        }
      }
    }
  }
}

// poor man's printf.
void print_hex(int i) {
  if (i > 0xf) {
    print_hex(i >> 4);
  }
  int c = i & 0xF;
  if (c < 10) {
    c += '0';
  } else {
    c = c - 10 + 'a';
  }
  putchar(c);
}


void draw_gun(int x, int y) {
  /* Left square */
  set_pixel(old_img_addr, x +  0, y + 5, 1);
  set_pixel(old_img_addr, x +  0, y + 6, 1);
  set_pixel(old_img_addr, x +  1, y + 5, 1);
  set_pixel(old_img_addr, x +  1, y + 6, 1);

  set_pixel(old_img_addr, x + 11, y + 5, 1);
  set_pixel(old_img_addr, x + 11, y + 6, 1);
  set_pixel(old_img_addr, x + 11, y + 7, 1);
  set_pixel(old_img_addr, x + 12, y + 4, 1);
  set_pixel(old_img_addr, x + 12, y + 8, 1);
  set_pixel(old_img_addr, x + 13, y + 3, 1);
  set_pixel(old_img_addr, x + 13, y + 9, 1);
  set_pixel(old_img_addr, x + 14, y + 3, 1);
  set_pixel(old_img_addr, x + 14, y + 9, 1);

  set_pixel(old_img_addr, x + 15, y + 6, 1);
  set_pixel(old_img_addr, x + 16, y + 4, 1);
  set_pixel(old_img_addr, x + 16, y + 8, 1);
  set_pixel(old_img_addr, x + 17, y + 5, 1);
  set_pixel(old_img_addr, x + 17, y + 6, 1);
  set_pixel(old_img_addr, x + 17, y + 7, 1);
  set_pixel(old_img_addr, x + 18, y + 6, 1);

  set_pixel(old_img_addr, x + 21, y + 3, 1);
  set_pixel(old_img_addr, x + 21, y + 4, 1);
  set_pixel(old_img_addr, x + 21, y + 5, 1);
  set_pixel(old_img_addr, x + 22, y + 3, 1);
  set_pixel(old_img_addr, x + 22, y + 4, 1);
  set_pixel(old_img_addr, x + 22, y + 5, 1);
  set_pixel(old_img_addr, x + 23, y + 2, 1);
  set_pixel(old_img_addr, x + 23, y + 6, 1);

  set_pixel(old_img_addr, x + 25, y + 1, 1);
  set_pixel(old_img_addr, x + 25, y + 2, 1);
  set_pixel(old_img_addr, x + 25, y + 6, 1);
  set_pixel(old_img_addr, x + 25, y + 7, 1);

  /* Right square */
  set_pixel(old_img_addr, x + 35, y + 3, 1);
  set_pixel(old_img_addr, x + 35, y + 4, 1);
  set_pixel(old_img_addr, x + 36, y + 3, 1);
  set_pixel(old_img_addr, x + 36, y + 4, 1);
}

int __start() {
  /* init_intc(); */
  write_mem(INTC_BASEADDR + INTC_IER,
            BIT(SENSOR_IRQ)|BIT(TIMER_IRQ));

  if (get_cpu_id() > 0)  {
    int i;
    for (i = 0; i <= get_cpu_id() << 8; i++)
      continue;
    printf("** SLAVE CPU BOOT COMPLETE **\n");
    printf("ID = ");
    print_hex(get_cpu_id());
    printf(", NB_SLAVE = ");
    print_hex(NB_SLAVE);
    printf("\n");
    printf("stack = ");
    print_hex((int)&i);
    printf("\n");
    should_redraw[get_cpu_id()] = 0;
    while (1) {
      while (should_redraw[get_cpu_id()] != 1) {
        // CPU1 wait after CPU0's call
        wait_for_irq();
      }
				
      printf("** SLAVE CPU redraws slice ");
      print_hex(get_cpu_id());
      printf(" **\n");
      game_of_life(get_cpu_id()); // On met à jour la moitié droite
      should_redraw[get_cpu_id()] = 0;
      printf("** SLAVE CPU completed redraw ");
      print_hex(get_cpu_id());
      printf(" **\n");
    }
  }
	
  // CPU0 initiate screen and wait event (normal loop)
  printf("-- boot complete -- \r\n");

  refresh = 0;
  old_img_addr = SRAM_BASEADDR;
  new_img_addr = SRAM_BASEADDR + 0x80000;

  clr_screen(old_img_addr);
  clr_screen(new_img_addr);
	
  draw_gun(10, 50);

  /* blinker */
  set_pixel(old_img_addr, 20, 10, 1);
  set_pixel(old_img_addr, 20, 11, 1);
  set_pixel(old_img_addr, 20, 12, 1);
   
  /* another blinker */
  set_pixel(old_img_addr, 30, 10, 1);
  set_pixel(old_img_addr, 31, 10, 1);
  set_pixel(old_img_addr, 32, 10, 1);

  /* A blinker at the borders between slices */
  {
    int i, j;
    for (i = 0; i <= NB_SLICES; i++)
      for (j = 1; j <= 3; j++)
        set_pixel(old_img_addr, i*SLICE_WIDTH, j, 1);
  }

  /* start vga */
  write_mem(VGA_BASEADDR + VGA_CFG_OFFSET, old_img_addr);

  /* program timer: 0x01000000 / (50MHz) = 0.33554432 seconds */
  write_mem(TIMER_BASEADDR + TIMER_0_TLR_OFFSET, 50000000); // 1sec
  write_mem(TIMER_BASEADDR + TIMER_0_CSR_OFFSET, TIMER_RELOAD);
  write_mem(TIMER_BASEADDR + TIMER_0_CSR_OFFSET, TIMER_START);

  /* program gpio to input */
  /* write_mem(GPIO_BASEADDR + GPIO_TRI_OFFSET, GPIO_INPUT); */

  while (1) {
    /* glider */
    set_pixel(old_img_addr, 11, 10, 1);
    set_pixel(old_img_addr, 12, 11, 1);
    set_pixel(old_img_addr, 10, 12, 1);
    set_pixel(old_img_addr, 11, 12, 1);
    set_pixel(old_img_addr, 12, 12, 1);

    /* another glider */
    set_pixel(old_img_addr, 51, 10, 1);
    set_pixel(old_img_addr, 50, 11, 1);
    set_pixel(old_img_addr, 52, 12, 1);
    set_pixel(old_img_addr, 51, 12, 1);
    set_pixel(old_img_addr, 50, 12, 1);

    /* another glider in right screen side */
    set_pixel(old_img_addr, SW_VGA_WIDTH/2 + 51, 10, 1);
    set_pixel(old_img_addr, SW_VGA_WIDTH/2 + 50, 11, 1);
    set_pixel(old_img_addr, SW_VGA_WIDTH/2 + 52, 12, 1);
    set_pixel(old_img_addr, SW_VGA_WIDTH/2 + 51, 12, 1);
    set_pixel(old_img_addr, SW_VGA_WIDTH/2 + 50, 12, 1);

    set_pixel(old_img_addr, 1, 31, 1);
    set_pixel(old_img_addr, 2, 31, 1);
    set_pixel(old_img_addr, 3, 31, 1);

    while (1) {
      /* START CUT */
      /* while (!compute) */
      /*   wait_for_irq(); */
      while (!(read_mem(TIMER_BASEADDR + TIMER_0_CSR_OFFSET) & TIMER_INTERRUPT))
        {
          /* printf("-- master CPU polls --\n"); */
        }
      
      printf("-- master CPU redraws --\n");
      game_of_life(get_cpu_id());
      printf("-- master CPU redraw DONE --\n");
      
      uint32_t tmp = old_img_addr;
      old_img_addr = new_img_addr;
      new_img_addr = tmp;
      refresh = 1;
      compute = 0;
      /* clear the timer interrupt */
      uint32_t timer_0_csr =
        read_mem(TIMER_BASEADDR + TIMER_0_CSR_OFFSET);
      write_mem(TIMER_BASEADDR + TIMER_0_CSR_OFFSET, timer_0_csr);
      /* ack_it(TIMER_IRQ); */

      /* write_mem(INTC_BASEADDR + INTC_IER, */
      /*           BIT(SENSOR_IRQ)|BIT(TIMER_IRQ)|BIT(VGA_IRQ)); */
      //cpu_relax();
      /* END CUT */

      /* uint32_t d = read_mem(GPIO_BASEADDR + GPIO_DATA_OFFSET); */
      /* if (TEST_BIT(d, GPIO_BTN0)) { */
      /*   break; */
      /* } */
    }
  }

  return 0;
}

void vga_isr() {
  if (refresh) {
    refresh = 0;
    printf("vga_isr: double buffer flip\r\n");
    write_mem(VGA_BASEADDR + VGA_CFG_OFFSET, old_img_addr);
    write_mem(INTC_BASEADDR + INTC_IER,
              BIT(SENSOR_IRQ)|BIT(TIMER_IRQ));
  }
}

void timer_0_isr() {
  int i;
  printf("timer_0_isr\r\n");
	
  /* for (i = 1; i <= NB_SLAVE; i++) { */
  /*   should_redraw[i] = 1; */
  /*   /\* purge_time(); *\/ */
  /*   write_mem(INTC_BASEADDR + (INTC_SIZE * i) + INTC_SIR, TIMER_IRQ); */
  /* } */
  compute = 1;

}

void timer_1_isr() {
  printf("timer_1_isr\n");
}

void __interrupt() {
  irq_received = 1;
  if (get_cpu_id() != 0)
    return;
  /* check if vga requested interrupt */
  uint32_t vga_irq = read_mem(VGA_BASEADDR + VGA_INT_OFFSET);
  if (vga_irq) {
    /* call the service routine */
    vga_isr();
    /* clear interrupt */
    write_mem(VGA_BASEADDR + VGA_INT_OFFSET, vga_irq);
    ack_it(VGA_IRQ);
  }

  /* check if timer 0 requested interrupt */
  uint32_t timer_0_csr =
    read_mem(TIMER_BASEADDR + TIMER_0_CSR_OFFSET);
  if (timer_0_csr & TIMER_INTERRUPT) {
    /* call the service routine */
    timer_0_isr();
  }

  uint32_t timer_1_csr =
    read_mem(TIMER_BASEADDR + TIMER_1_CSR_OFFSET);
  if (timer_1_csr & TIMER_INTERRUPT) {
    /* call the service routine */
    timer_1_isr();
    /* clear the timer interrupt */
    write_mem(TIMER_BASEADDR + TIMER_1_CSR_OFFSET, timer_1_csr);
    ack_it(TIMER_IRQ);
  }

  /* check if uart requested interrupt */
  /* FIXME: no uart in TLM platform */
}
